import copy
import typing
from dataclasses import is_dataclass

_marker = object()


def drop_empty(schema, data):
    result = {}
    for attr, prop in sorted(schema.__dataclass_fields__.items(), key=lambda x: x[0]):
        if data.get(attr, _marker) is _marker:
            if prop.metadata.get("exclude_if_empty", None) is True:
                continue
        result[attr] = copy.deepcopy(data[attr])
    return result


def dataclass_get_type(field, exclude_if_empty=False):
    metadata = {
        "required": _marker,
        "exclude_if_empty": exclude_if_empty,
        "validators": [],
    }

    metadata.update(field.metadata)

    origin = getattr(field.type, "__origin__", None)
    required = True
    if origin == typing.Union:
        if len(field.type.__args__) == 2:
            if field.type.__args__[1] == type(None):
                required = False
            typ = field.type.__args__[0]
    else:
        typ = field.type

    if metadata["required"] is _marker:
        metadata["required"] = required

    if field.metadata.get("required", None) is not None:
        metadata["required"] = field.metadata["required"]

    required = metadata["required"]

    origin = getattr(typ, "__origin__", None)

    if origin == list:
        if getattr(typ, "__args__", None):
            return {
                "name": field.name,
                "type": list,
                "schema": field.type.__args__[0],
                "required": required,
                "metadata": metadata,
            }
        else:
            return {
                "name": field.name,
                "type": list,
                "required": required,
                "metadata": metadata,
            }

    return {"type": typ, "required": required, "metadata": metadata}


def is_dataclass_field(field):
    t = dataclass_get_type(field)

    if is_dataclass(t["type"]):
        return t

    return None


def dataclass_check_type(field, basetype):

    t = dataclass_get_type(field)

    if t["type"] == basetype:
        return t

    # wtf bool is a subclass of integer?
    if t["type"] == bool and basetype == int:
        return None

    if issubclass(t["type"], basetype):
        return t

    return None



import dataclasses
import typing
from datetime import date, datetime, timedelta

import colander
import pytz

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field
from .dc2colander import Mapping, SchemaNode, colander_params
from .dc2colander import dataclass_field_to_colander_schemanode as orig_dc2colander_node
from .dc2colander import dc2colander

epoch_date = date(1970, 1, 1)


class Boolean(colander.Boolean):
    def deserialize(self, node, appstruct):
        if appstruct is True:
            appstruct = "true"
        elif appstruct is False:
            appstruct = "false"
        elif appstruct is None:
            appstruct = colander.null
        return super(Boolean, self).deserialize(node, appstruct)

    def serialize(self, node, appstruct):
        result = super(Boolean, self).serialize(node, appstruct)
        if result is colander.null:
            return None
        if result is not colander.null:
            if result.lower() == "true":
                result = True
            else:
                result = False
        return result


class Float(colander.Float):
    def deserialize(self, node, cstruct):
        if cstruct is None:
            return colander.null
        if cstruct is not colander.null:
            cstruct = float(cstruct)
        return super(Float, self).deserialize(node, cstruct)

    def serialize(self, node, appstruct):
        result = super(Float, self).serialize(node, appstruct)
        if result is colander.null:
            return None
        if result is not colander.null:
            result = float(result)
        return result


class Int(colander.Int):
    def deserialize(self, node, cstruct):
        if cstruct is None:
            return colander.null

        if cstruct is not colander.null:
            cstruct = int(cstruct)
        return super(Int, self).deserialize(node, cstruct)

    def serialize(self, node, appstruct):
        result = super(Int, self).serialize(node, appstruct)
        if result is colander.null:
            return None

        if result is not colander.null:
            result = int(result)
        return result


class Str(colander.Str):
    def deserialize(self, node, cstruct):
        if cstruct is None:
            return colander.null
        return super(Str, self).deserialize(node, cstruct)

    def serialize(self, node, appstruct):
        if appstruct is None:
            return None
        result = super(Str, self).serialize(node, appstruct)
        if result is colander.null:
            return None
        return result


class Date(colander.Date):
    def serialize(self, node, appstruct):
        result = super(Date, self).serialize(node, appstruct)
        if result is colander.null:
            return None

        return (appstruct - epoch_date).days

    def deserialize(self, node, cstruct):
        if cstruct and not (isinstance(cstruct, int) or isinstance(cstruct, str)):
            raise colander.Invalid(
                node,
                (
                    "Date is expected to be number of days after 1970-01-01, "
                    "or in ISO formatted date string"
                ),
            )

        if cstruct and isinstance(cstruct, int):
            cstruct = (epoch_date + timedelta(days=cstruct)).strftime(r"%Y-%m-%d")
        return super().deserialize(node, cstruct)


class DateTime(colander.DateTime):
    def serialize(self, node, appstruct):
        if appstruct:
            appstruct = appstruct.astimezone(pytz.UTC)
        result = super(DateTime, self).serialize(node, appstruct)
        if result is colander.null:
            return None

        return int(appstruct.timestamp() * 1000)

    def deserialize(self, node, cstruct):
        if cstruct and not (isinstance(cstruct, int) or isinstance(cstruct, str)):
            raise colander.Invalid(
                node,
                (
                    "DateTime is expected to in Unix timestamp "
                    "in miliseconds in UTC, or in ISO formatted date string"
                ),
            )
        if cstruct and isinstance(cstruct, int):
            cstruct = datetime.fromtimestamp(
                int(cstruct) / 1000, tz=pytz.UTC
            ).isoformat()
        result = super().deserialize(node, cstruct)
        if result:
            return result.astimezone(self.default_tzinfo)
        return result


def dataclass_field_to_colander_schemanode(
    prop: dataclasses.Field,
    schema,
    request,
    oid_prefix="deformField",
    mode=None,
    metadata=None,
    default_tzinfo=pytz.UTC,
) -> colander.SchemaNode:

    t = dataclass_get_type(prop)
    metadata = metadata or {}
    t["metadata"].update(metadata)
    json_field_factory = t["metadata"].get("colanderjson.field_factory", None)
    if json_field_factory:
        params = colander_params(
            prop,
            oid_prefix,
            typ=json_field_factory(request),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    field_factory = t["metadata"].get("colander.field_factory", None)
    if field_factory:
        params = colander_params(
            prop,
            oid_prefix,
            typ=field_factory(request),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == date:
        params = colander_params(
            prop, oid_prefix, typ=Date(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == datetime:
        params = colander_params(
            prop,
            oid_prefix,
            typ=DateTime(default_tzinfo=default_tzinfo),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == str:
        params = colander_params(
            prop,
            oid_prefix,
            typ=Str(allow_empty=True),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == int:
        params = colander_params(
            prop, oid_prefix, typ=Int(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == float:
        params = colander_params(
            prop, oid_prefix, typ=Float(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == bool:
        params = colander_params(
            prop, oid_prefix, typ=Boolean(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)

    if is_dataclass_field(prop):
        subtype = dc2colanderjson(
            t["type"],
            colander_schema_type=colander.MappingSchema,
            request=request,
            mode=mode,
            default_tzinfo=default_tzinfo,
            field_metadata=metadata,
        )
        return subtype()

    if t["type"] == dict:
        params = colander_params(
            prop,
            oid_prefix,
            typ=Mapping(unknown="preserve"),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)

    return orig_dc2colander_node(
        prop=prop,
        schema=schema,
        request=request,
        oid_prefix=oid_prefix,
        mode=mode,
        default_tzinfo=default_tzinfo,
        metadata=metadata,
    )


def dc2colanderjson(
    schema,
    *,
    include_fields: typing.List[str] = None,
    exclude_fields: typing.List[str] = None,
    hidden_fields: typing.List[str] = None,
    readonly_fields: typing.List[str] = None,
    include_schema_validators: bool = True,
    colander_schema_type: typing.Type[colander.Schema] = colander.MappingSchema,
    oid_prefix: str = "deformField",
    request=None,
    mode="default",
    default_tzinfo=pytz.UTC,
    field_metadata=None,
) -> typing.Type[colander.MappingSchema]:

    """
    Converts ``dataclass`` to ``colander.Schema`` that serializes to JSON.

    - date is serialized as number days from epoch
    - datetime is serialized as number of miliseconds from epoch

    Accepted parameters are the same as ``inverter.dc2colander.convert``.
    """
    return dc2colander(
        schema,
        request=request,
        include_fields=include_fields,
        exclude_fields=exclude_fields,
        hidden_fields=hidden_fields,
        readonly_fields=readonly_fields,
        include_schema_validators=include_schema_validators,
        colander_schema_type=colander_schema_type,
        oid_prefix=oid_prefix,
        dataclass_field_to_colander_schemanode=dataclass_field_to_colander_schemanode,
        mode=mode,
        field_metadata=field_metadata,
        default_tzinfo=default_tzinfo,
    )


convert = dc2colanderjson



import dataclasses
import json
import typing
from datetime import date, datetime, timedelta

import colander
import pytz

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field
from .dc2colander import SchemaNode, colander_params
from .dc2colander import dataclass_field_to_colander_schemanode as orig_dc2colander_node
from .dc2colander import dc2colander
from .dc2colanderjson import Boolean, Float, Int, Str


class Date(colander.Date):
    def serialize(self, node, appstruct):
        result = super(Date, self).serialize(node, appstruct)
        if result is colander.null:
            return None
        return result

    def deserialize(self, node, cstruct):
        return super().deserialize(node, cstruct)


class DateTime(colander.DateTime):
    def serialize(self, node, appstruct):
        if appstruct:
            appstruct = appstruct.astimezone(pytz.UTC)
        result = super(DateTime, self).serialize(node, appstruct)
        if result is colander.null:
            return None
        return result

    def deserialize(self, node, cstruct):
        res = super().deserialize(node, cstruct)
        if res:
            res = res.astimezone(self.default_tzinfo)
        return res


def dataclass_field_to_colander_schemanode(
    prop: dataclasses.Field,
    schema,
    request,
    oid_prefix="deformField",
    mode=None,
    default_tzinfo=pytz.UTC,
    metadata=None,
) -> colander.SchemaNode:

    t = dataclass_get_type(prop)
    field_factory = t["metadata"].get("colander.field_factory", None)
    if field_factory:
        params = colander_params(
            prop,
            oid_prefix,
            typ=field_factory(request),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)

    if t["type"] == date:
        params = colander_params(
            prop, oid_prefix, typ=Date(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == datetime:
        params = colander_params(
            prop,
            oid_prefix,
            typ=DateTime(default_tzinfo=default_tzinfo),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == str:
        params = colander_params(
            prop, oid_prefix, typ=Str(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == int:
        params = colander_params(
            prop, oid_prefix, typ=Int(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == float:
        params = colander_params(
            prop, oid_prefix, typ=Float(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)
    if t["type"] == bool:
        params = colander_params(
            prop, oid_prefix, typ=Boolean(), schema=schema, request=request, mode=mode
        )
        return SchemaNode(**params)

    if is_dataclass_field(prop):
        subtype = dc2colanderESjson(
            prop,
            colander_schema_type=colander.MappingSchema,
            request=request,
            mode=mode,
            field_metadata=metadata,
            default_tzinfo=default_tzinfo,
        )
        return subtype()

    if t["type"] == dict:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.Mapping(unknown="preserve"),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)

    return orig_dc2colander_node(
        prop=prop,
        schema=schema,
        request=request,
        oid_prefix=oid_prefix,
        mode=mode,
        default_tzinfo=default_tzinfo,
        metadata=metadata,
    )


def dc2colanderESjson(
    schema,
    *,
    include_fields: typing.List[str] = None,
    exclude_fields: typing.List[str] = None,
    hidden_fields: typing.List[str] = None,
    readonly_fields: typing.List[str] = None,
    include_schema_validators: bool = True,
    colander_schema_type: typing.Type[colander.Schema] = colander.MappingSchema,
    oid_prefix: str = "deformField",
    request=None,
    default_tzinfo=pytz.UTC,
    mode="default",
    field_metadata=None,
) -> typing.Type[colander.MappingSchema]:
    """
    Converts ``dataclass`` to ``colander.Schema`` that serializes to ElasticSearch
    compatible dictionary.

    - date is serialized as YYYY-MM-DD string
    - datetime is serialized as iso8601 string

    Accepted parameters are the same as ``inverter.dc2colander.convert``.
    """
    return dc2colander(
        schema,
        request=request,
        include_fields=include_fields,
        exclude_fields=exclude_fields,
        hidden_fields=hidden_fields,
        readonly_fields=readonly_fields,
        include_schema_validators=include_schema_validators,
        colander_schema_type=colander_schema_type,
        oid_prefix=oid_prefix,
        dataclass_field_to_colander_schemanode=dataclass_field_to_colander_schemanode,
        mode=mode,
        default_tzinfo=default_tzinfo,
        field_metadata=field_metadata,
    )


convert = dc2colanderESjson



import datetime
import typing

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field


def dataclass_field_to_avsc_field(prop, schema, request, ignore_required=False):
    """
    Converts ``dataclass.Field`` to Avro schema field dictionary.

    :param prop: ``dataclass.Field`` object
    :param schema: ``dataclass`` class
    :param request: request object, accepts any.
    :param ignore_required: if ``True``, force all fields as non-required
    :type ignore_required: bool
    """

    t = dataclass_get_type(prop)
    field = {"name": prop.name}

    if not ignore_required:
        required = prop.metadata.get("required", False)
    else:
        required = False

    if t["type"] == str and prop.metadata.get("format", None) == "uuid":
        field["type"] = [{"type": "string", "logicalType": "uuid"}]
        if not required:
            field["type"].append("null")
        return field

    if t["type"] == str:
        field["type"] = ["string"]
        if not required:
            field["type"].append("null")
        return field

    if t["type"] == int:
        field["type"] = ["int"]
        if not required:
            field["type"].append("null")
        return field

    if t["type"] == float:
        field["type"] = ["double"]
        if not required:
            field["type"].append("null")
        return field

    if t["type"] == bool:
        field["type"] = ["boolean"]
        if not required:
            field["type"].append("null")
        return field

    if t["type"] == datetime.datetime:
        field["type"] = [
            {
                "type": "long",
                "logicalType": "timestamp-millis",
            },
        ]
        if not required:
            field["type"].append("null")
        return field

    if t["type"] == datetime.date:
        field["type"] = [
            {
                "type": "int",
                "logicalType": "date",
            }
        ]
        if not required:
            field["type"].append("null")
        return field

    if is_dataclass_field(prop):
        subtype = dc2avsc(prop, request=request)
        return subtype

    if t["type"] == dict:
        # FIXME: dictionary field are expected to be
        # encoded as string for now.
        # ideally it should be schemaless dict, but
        # it is unclear how to achieve this with
        # avro
        field["type"] = ["string"]
        if not required:
            field["type"].append("null")
        return field
    raise TypeError("Unknown Avro type for %s" % t["type"])


def dc2avsc(
    schema,
    *,
    request=None,
    include_fields: typing.List[str] = None,
    exclude_fields: typing.List[str] = None,
    namespace="inverter",
    ignore_required=True,
):
    """
    Converts ``dataclass`` to Avro Schema JSON dictionary

    :param schema: ``dataclass`` class
    :param request: request object, accepts Any
    :param include_fields: List of field names to include
    :type include_fields: typing.List[str]
    :param exclude_fields: List of field names to exclude
    :type exclude_fields: typing.List[str]
    :param namespace: Avro schema namespace, defaults to 'inverter'
    :param ignore_required: if True, force all fields to be non-required
    :type ignore_required: bool

    :return: dictionary representing Avro Schema.
    """
    result = {
        "namespace": namespace,
        "type": "record",
        "name": str(schema.__name__),
        "fields": [],
    }
    for attr, prop in schema.__dataclass_fields__.items():
        field = dataclass_field_to_avsc_field(
            prop, schema=schema, request=request, ignore_required=ignore_required
        )
        result["fields"].append(field)

    return result


convert = dc2avsc



import jsonobject
import jsl


def jsl_field_to_jsonobject_property(prop: jsl.BaseField) -> jsonobject.JsonProperty:
    if isinstance(prop, jsl.DateTimeField):
        return jsonobject.DateTimeProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.StringField):
        return jsonobject.StringProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.IntField):
        return jsonobject.IntegerProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.DictField):
        return jsonobject.DictProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.NumberField):
        return jsonobject.FloatProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.BooleanField):
        return jsonobject.BooleanProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.DocumentField):
        if prop.document_cls:
            subtype = jsl2jsonobject(prop.document_cls)
            return jsonobject.DictProperty(
                name=prop.name, item_type=subtype, required=prop.required
            )
        return jsonobject.DictProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.ArrayField):
        if prop.items:
            if isinstance(prop.items, jsl.DocumentField):
                subtype = jsl2jsonobject(prop.items.document_cls)
            elif isinstance(prop.items, jsl.BaseField):
                subtype = jsl_field_to_jsonobject_property(prop.items)
            else:
                raise KeyError(prop.items)
            return jsonobject.ListProperty(item_type=subtype, required=prop.required)
        return jsonobject.ListProperty(name=prop.name, required=prop.required)

    raise KeyError(prop)


def jsl2jsonobject(schema):
    # output jsonobject schema from jsl schema
    attrs = {}
    for attr, prop in schema._fields.items():
        prop.name = attr
        attrs[attr] = jsl_field_to_jsonobject_property(prop)

    Schema = type("Schema", (jsonobject.JsonObject,), attrs)

    return Schema

convert = jsl2jsonobject



import copy
import dataclasses
import typing
from datetime import date, datetime

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field


def dataclass_field_to_esmapping(
    prop: dataclasses.Field, schema, request, *, metadata=None
):
    t = dataclass_get_type(prop)
    metadata = metadata or {}
    meta = copy.deepcopy(t["metadata"])
    meta.update(metadata)
    index = meta.get("index", None)
    mapping_opts = meta.get("es.mapping_options", {})
    if index is not None:
        mapping_opts.setdefault("index", index)
    if t["type"] == date:
        mfield = {"type": "date"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == datetime:
        mfield = {"type": "date"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == str:
        fmt = meta.get("format", None)
        if fmt is None:
            mfield = {"type": "keyword"}
        elif fmt == "text" or fmt.startswith("text/"):
            mfield = {"type": "text", "fields": {"raw": {"type": "keyword"}}}
        else:
            mfield = {"type": "keyword"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == int:
        mfield = {"type": "long"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == float:
        mfield = {"type": "double"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == bool:
        mfield = {"type": "boolean"}
        mfield.update(mapping_opts)
        return mfield
    if is_dataclass_field(prop):
        mfield = {"type": "object"}
        mfield.update(mapping_opts)
    if t["type"] == dict:
        mfield = {"type": "object"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == list:
        mfield = {"type": "nested"}
        mfield.update(mapping_opts)
        return mfield
    if t["type"] == set:
        mfield = {"type": "nested"}
        mfield.update(mapping_opts)
        return mfield
    raise KeyError(prop)


def dc2esmapping(
    schema: type,
    *,
    request: typing.Any = None,
    metadata: typing.Optional[dict] = None,
    include_fields=None,
    exclude_fields=None,
):

    include_fields = include_fields or []
    exclude_fields = exclude_fields or []
    mprops = {}
    if include_fields:
        for attr, prop in schema.__dataclass_fields__.items():
            if prop.name in include_fields and prop.name not in exclude_fields:
                mprops[attr] = dataclass_field_to_esmapping(
                    prop, schema, request, metadata=metadata
                )
    else:
        for attr, prop in schema.__dataclass_fields__.items():
            if prop.name not in exclude_fields:
                mprops[attr] = dataclass_field_to_esmapping(
                    prop, schema, request, metadata=metadata
                )

    return {"mappings": {"properties": mprops}}


convert = dc2esmapping



import copy
import dataclasses
import typing
from dataclasses import field
from datetime import date, datetime
from importlib import import_module

import colander
import sqlalchemy
import sqlalchemy_jsonfield as sajson
import sqlalchemy_utils as sautils
from deform.widget import HiddenWidget
from pkg_resources import resource_filename

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field


def sqlalchemy_params(prop, typ, **kwargs):
    t = dataclass_get_type(prop)

    params = {"name": prop.name, "type_": typ}

    if not isinstance(prop.default, dataclasses._MISSING_TYPE):
        params["default"] = prop.default

    if not isinstance(prop.default_factory, dataclasses._MISSING_TYPE):
        params["default"] = prop.default_factory

    if t["metadata"].get("primary_key", None) is True:
        params["primary_key"] = True

    if t["metadata"].get("index", None) is True:
        params["index"] = True

    if t["metadata"].get("autoincrement", None) is True:
        params["autoincrement"] = True

    if t["metadata"].get("unique", None) is True:
        params["unique"] = True

    params.update(kwargs)
    return params


def dataclass_field_to_sqla_col(prop: dataclasses.Field) -> sqlalchemy.Column:
    t = dataclass_get_type(prop)
    if t["type"] == date:
        params = sqlalchemy_params(prop, typ=sqlalchemy.Date())
        return sqlalchemy.Column(**params)
    if t["type"] == datetime:
        params = sqlalchemy_params(prop, typ=sqlalchemy.DateTime(timezone=True))
        return sqlalchemy.Column(**params)
    if t["type"] == str:
        str_format = t["metadata"].get("format", None)

        if str_format and "/" in str_format:
            str_format = str_format.split("/")[0]

        if str_format == "text":
            params = sqlalchemy_params(prop, typ=sqlalchemy.Text())
        elif str_format == "uuid":
            params = sqlalchemy_params(prop, typ=sautils.UUIDType())
        elif str_format == "fulltextindex":
            params = sqlalchemy_params(prop, typ=sautils.TSVectorType)
        else:
            str_len = prop.metadata.get("length", 256)
            params = sqlalchemy_params(prop, typ=sqlalchemy.String(str_len))
        return sqlalchemy.Column(**params)
    if t["type"] == int:
        if t["metadata"].get("format", None) == "bigint":
            params = sqlalchemy_params(prop, typ=sqlalchemy.BigInteger())
        else:
            params = sqlalchemy_params(prop, typ=sqlalchemy.Integer())
        return sqlalchemy.Column(**params)
    if t["type"] == float:
        if t["metadata"].get("format", None) == "numeric":
            params = sqlalchemy_params(prop, typ=sqlalchemy.Numeric())
        else:
            params = sqlalchemy_params(prop, typ=sqlalchemy.Float())

        return sqlalchemy.Column(**params)
    if t["type"] == bool:
        params = sqlalchemy_params(prop, typ=sqlalchemy.Boolean())
        return sqlalchemy.Column(**params)

    if is_dataclass_field(prop):
        raise NotImplementedError("Sub schema is not supported")

    if t["type"] == dict:
        params = sqlalchemy_params(prop, typ=sajson.JSONField())
        return sqlalchemy.Column(**params)
    if t["type"] == list:
        params = sqlalchemy_params(prop, typ=sajson.JSONField())
        return sqlalchemy.Column(**params)

    raise KeyError(prop)


def dc2pgsqla(schema, metadata, *, name=None) -> sqlalchemy.Table:
    """
    Convert ``dataclass`` to ``sqlalchemy`` ORM model

    :param schema: ``dataclass`` class
    :param metadata: ``sqlalchemy.MetaData`` object
    :param name: model name

    :return: ``sqlalchemy`` ORM class

    **Field metadata handling**

    - ``primary_key: bool`` - primary key flag
    - ``index: bool`` - flag on whether to index the column
    - ``autoincrement: bool`` - flag on whether to make column autoincrement
    - ``unique: bool`` - flag on whether to make column unique
    - ``searchable: bool`` - flag on whether to make column searchable using PGSQL Trigram
    - ``format: str`` - format of data: ``uuid``, ``text``, ``fulltextindex``,
      ``bigint``, ``numeric``. This forces SQLAlchemy to use specific data type
      for the format.
    """
    if name is None:
        if getattr(schema, "__table_name__", None):
            name = schema.__table_name__
        else:
            name = schema.__name__.lower()

    cols = []

    for attr, prop in sorted(schema.__dataclass_fields__.items(), key=lambda x: x[0]):
        prop = dataclass_field_to_sqla_col(prop)
        cols.append(prop)

    Table = sqlalchemy.Table(name, metadata, *cols, extend_existing=True)

    for attr, prop in schema.__dataclass_fields__.items():
        if prop.type in [str, typing.Optional[str]] and prop.metadata.get(
            "searchable", None
        ):
            sqlalchemy.Index(
                "ix_%s_%s_trgm_search" % (name, attr),
                getattr(Table.c, attr),
                postgresql_ops={attr: "gin_trgm_ops"},
                postgresql_using="gin",
                unique=False,
            )

    # FIXME: reject nested schema

    return Table


convert = dc2pgsqla



import dataclasses
import typing
from datetime import date, datetime

import jsl

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field


def _set_nullable(prop):
    if not prop.required:
        return jsl.OneOfField([prop, jsl.NullField(name=prop.name)])
    return prop


def dc2jsl(
    schema, *, ignore_required=False, additional_properties=False, mode="default"
):
    """
    Convert ``dataclass`` to ``jsl`` JSON Schema.

    :param schema: ``dataclass`` class
    :param ignore_required: if ``True``, set all fields as nullable
    :param additional_properties: Allow ``additional_properties`` in JSON Schema
    :param mode: mode flag

    :return: ``jsl.Document`` class
    """

    nullable = ignore_required
    attrs = {}

    _additional_properties = additional_properties

    class Options(object):
        additional_properties = _additional_properties

    for attr, prop in schema.__dataclass_fields__.items():
        prop = dataclass_field_to_jsl_field(prop, nullable=nullable, mode=mode)
        if nullable:
            attrs[attr] = _set_nullable(prop)
        else:
            if not prop.required:
                attrs[attr] = prop
            else:
                if isinstance(prop, jsl.StringField):
                    if not prop.pattern:
                        prop.pattern = ".+"
                attrs[attr] = prop

    attrs["Options"] = Options
    Schema = type("Schema", (jsl.Document,), attrs)

    return Schema


def dataclass_field_to_jsl_field(
    prop: dataclasses.Field, nullable=False, mode="default"
) -> jsl.BaseField:

    if mode in ["edit", "edit-process"]:
        update_mode = True
    else:
        update_mode = False
    t = dataclass_check_type(prop, date)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.DateTimeField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, datetime)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.DateTimeField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, str)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.StringField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, int)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.IntField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, float)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.NumberField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, bool)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.BooleanField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, dict)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.DictField(name=prop.name, required=t["required"])

    t = is_dataclass_field(prop)
    if t:
        if update_mode:
            t["required"] = False
        subtype = dc2jsl(t["schema"], nullable=nullable)
        return jsl.DocumentField(
            name=prop.name, document_cls=subtype, required=t["required"]
        )

    t = dataclass_check_type(prop, list)
    if t:
        if update_mode:
            t["required"] = False
        return jsl.ArrayField(name=prop.name, required=t["required"])

    t = dataclass_check_type(prop, typing.List)
    if t:
        if update_mode:
            t["required"] = False

        if "schema" not in t.keys():
            return jsl.ArrayField(name=prop.name, required=t["required"])

        if dataclasses.is_dataclass(t["schema"]):
            subtype = jsl.DocumentField(
                document_cls=dc2jsl(t["schema"], nullable=nullable)
            )
        elif t["schema"] == str:
            subtype = jsl.StringField(name=prop.name)
        elif t["schema"] == int:
            subtype = jsl.IntField(name=prop.name)
        elif t["schema"] == float:
            subtype = jsl.NumberField(name=prop.name)
        elif t["schema"] == dict:
            subtype = jsl.DictField(name=prop.name)
        else:
            raise KeyError(t["schema"])
        return jsl.ArrayField(items=subtype, required=t["required"])

    raise KeyError(prop)


convert = dc2jsl



import jsonobject
import jsl


def _set_nullable(prop):
    if not prop.required:
        return jsl.OneOfField([prop, jsl.NullField(name=prop.name)])
    return prop


def jsonobject2jsl(schema, nullable=False):
    # output jsl schema from jsonobject schema
    attrs = {}

    class Options(object):
        additional_properties = True

    for attr, prop in schema._properties_by_attr.items():
        prop = jsonobject_property_to_jsl_field(prop, nullable=nullable)
        if nullable:
            attrs[attr] = _set_nullable(prop)
        else:
            if not prop.required:
                attrs[attr] = prop
            else:
                if isinstance(prop, jsl.StringField):
                    if not prop.pattern:
                        prop.pattern = ".+"
                attrs[attr] = prop

    attrs["Options"] = Options
    Schema = type("Schema", (jsl.Document,), attrs)

    return Schema


def jsonobject_property_to_jsl_field(
    prop: jsonobject.JsonProperty, nullable=False
) -> jsl.BaseField:
    if isinstance(prop, jsonobject.DateProperty):
        return jsl.DateTimeField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.DateTimeProperty):
        return jsl.DateTimeField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.StringProperty):
        return jsl.StringField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.IntegerProperty):
        return jsl.IntField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.FloatProperty):
        return jsl.NumberField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.BooleanProperty):
        return jsl.BooleanField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.DictProperty):
        if prop.item_wrapper:
            subtype = jsonobject2jsl(prop.item_wrapper.item_type, nullable=nullable)
            return jsl.DocumentField(
                name=prop.name, document_cls=subtype, required=prop.required
            )
        return jsl.DictField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.ListProperty):
        if prop.item_wrapper:
            if isinstance(prop.item_wrapper, jsonobject.ObjectProperty):
                if issubclass(prop.item_wrapper.item_type, jsonobject.JsonObject):
                    subtype = jsl.DocumentField(
                        document_cls=jsonobject2jsl(prop.item_wrapper.item_type),
                        nullable=nullable,
                    )
                elif isinstance(prop.item_wrapper.item_type, jsonobject.JsonProperty):
                    subtype = jsonobject_property_to_jsl_field(
                        prop.item_wrapper.item_type
                    )
                else:
                    raise KeyError(prop.item_wrapper.item_type)
            elif isinstance(prop.item_wrapper, jsonobject.StringProperty):
                subtype = jsl.StringField(name=prop.name)
            elif isinstance(prop.item_wrapper, jsonobject.IntegerProperty):
                subtype = jsl.IntField(name=prop.name)
            elif isinstance(prop.item_wrapper, jsonobject.FloatProperty):
                subtype = jsl.NumberField(name=prop.name)
            elif isinstance(prop.item_wrapper, jsonobject.DictProperty):
                subtype = jsl.DictField(name=prop.name)
            else:
                raise KeyError(prop.item_wrapper)
            return jsl.ArrayField(items=subtype, required=prop.required)
        return jsl.ArrayField(name=prop.name, required=prop.required)

    raise KeyError(prop)

convert = jsonobject2jsl



from setuptools import setup, find_packages
import sys, os

version = '0.1.2'

def readfile(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        out = f.read()
    return out

desc = '\n'.join([readfile('README.rst'), readfile('CHANGELOG.rst')])

setup(name='inverter',
      version=version,
      description="Convert dataclass to another class",
      long_description=desc,
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
      ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='dataclass converter',
      author='Izhar Firdaus',
      author_email='kagesenshi.87@gmail.com',
      url='http://github.com/morpframework/inverter/',
      license='Apache-2.0',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'colander',
          'jsl',
          'sqlalchemy',
          'sqlalchemy_utils',
          'sqlalchemy_jsonfield',
          'deform',
          'pytz',
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



import copy
import dataclasses
import typing
from dataclasses import _MISSING_TYPE, field
from datetime import date, datetime
from importlib import import_module

import colander
import deform
import pytz
from deform.schema import default_widget_makers
from deform.widget import HiddenWidget, TextAreaWidget, TextInputWidget
from pkg_resources import resource_filename

from .common import dataclass_check_type, dataclass_get_type, is_dataclass_field


def replace_colander_null(appstruct, value=None):
    """
    Replace ``colander.null`` with default value

    :param appstruct: colander ``appstruct`` dictionary
    :type appstruct: ``dict``
    :param value: value to replace ``colander.null`` with, defaults to ``None``

    :return: ``appstruct`` dictionary that have been replaced
    :rtype: ``dict``
    """
    out = {}
    for k, v in appstruct.items():
        if isinstance(v, dict):
            out[k] = replace_colander_null(v)
        elif isinstance(v, list):
            out[k] = list(map(lambda x: x if x != colander.null else value, v))
        else:
            out[k] = v if v != colander.null else value
    return out


class ValidatorsWrapper(object):
    """
    This wrapper is used internally to validate data using multiple validators
    """

    def __init__(
        self,
        validators: typing.List[typing.Callable],
        request: typing.Any,
        schema,
        mode=None,
    ):
        """
        :param validators: list of validator callables that accept following parameters:
                           - ``request`` - request object
                           - ``schema`` - dataclass schema
                           - ``field`` - field name
                           - ``value`` - value to be validated
                           - ``mode`` - one of the following: ``'default'``, ``'edit'``, ``'edit-process'``

        :param request: request object that is passed to the schema converter. Accept any.
        :param schema: dataclass schema
        :param mode:  one of the following: ``default``, ``'edit'``, ``'edit-process'``

        """
        self.request = request
        self.schema = schema
        self.mode = mode
        self.validators = validators

    def __call__(self, node, value):
        """
        Execute validators and raise errors

        :param node: ``colander`` field node
        :param value: value to be validated

        :raises colander.Invalid: raised when invalid data is found
        """
        for validator in self.validators:
            error = validator(
                request=self.request,
                schema=self.schema,
                field=node.name,
                value=value,
                mode=self.mode,
            )
            if error:
                raise colander.Invalid(node, error)


class PreparersWrapper(object):
    """
    This wrapper is used internally to prepare data using multiple preparers
    """

    def __init__(
        self, preparers: typing.List[typing.Callable], request, schema, mode=None
    ):
        """
        :param preparers: list of validator callables that accept following parameters:
                           - ``request`` - request object
                           - ``schema`` - dataclass schema
                           - ``value`` - value to be validated
                           - ``mode`` - one of the following: ``'default'``, ``'edit'``, ``'edit-process'``

        :param request: request object that is passed to the schema converter. Accept any.
        :param schema: dataclass schema
        :param mode:  one of the following: ``'default'``, ``'edit'``, ``'edit-process'``

        """
        self.preparers = preparers
        self.request = request
        self.schema = schema
        self.mode = mode

    def __call__(self, value: typing.Any) -> typing.Any:
        """

        Execute preparers against ``value``

        :param value: data value
        :return: prepared value

        """
        if value is colander.null:
            value = None
        for preparer in self.preparers:
            value = preparer(
                request=self.request, schema=self.schema, value=value, mode=self.mode
            )
        return value


class SchemaNode(colander.SchemaNode):
    """
    Replace the way SchemaNode handles serialization

    :meta private:
    """

    def serialize(self, appstruct=colander.null):
        """
        If ``colander.drop`` is used, ``deform`` seems to output ``colander.drop``
        as the default value in the forms. This fixes that.

        :meta private:
        """
        # workaround with deform serialization issue with colander.drop
        if appstruct is colander.null:
            appstruct = self.default
        if appstruct is colander.drop:
            appstruct = colander.null
        # if isinstance(appstruct, colander.deferred):
        #    appstruct = colander.null
        cstruct = self.typ.serialize(self, appstruct)
        return cstruct


def colander_params(
    prop: dataclasses.Field,
    oid_prefix: str,
    schema: type,
    request: typing.Any,
    mode=None,
    **kwargs,
) -> dict:
    """
    Get parameters for ``colander.SchemaNode`` constructor from ``dataclass.Field``

    :param prop: ``dataclass.Field`` object
    :param oid_prefix: string prefix for use as field oid
    :param schema: ``dataclass`` based class
    :param request: ``request`` object. Accepts anything, as it is merely passed to validators
    :param mode: one of the following: ``'default'``, ``'edit'``, ``'edit-process'``
    :param kwargs: additional parameters to be included into output. Will override derived parameters.

    :return: dictionary of parameters for ``colander.SchemaNode``

    """
    t = dataclass_get_type(prop)

    default_value = None
    if t["type"] == dict:
        default_value = {}
    elif t["type"] == list:
        default_value = []
    elif t["type"] == set:
        default_value = set()

    if (
        not isinstance(prop.default, dataclasses._MISSING_TYPE)
        and prop.default is not None
    ):
        default_value = prop.default
    elif (
        not isinstance(prop.default_factory, dataclasses._MISSING_TYPE)
        and prop.default_factory is not None
    ):
        default_value = prop.default_factory()

    params = {
        "name": prop.name,
        "oid": "%s-%s" % (oid_prefix, prop.name),
        "missing": colander.required if t["required"] else default_value,
        "default": default_value,
    }

    if "deform.widget" in prop.metadata.keys():
        params["widget"] = copy.copy(prop.metadata["deform.widget"])
    elif "deform.widget_factory" in prop.metadata.keys():
        params["widget"] = prop.metadata["deform.widget_factory"](request)

    validators = prop.metadata.get("validators", None)
    if validators:
        params["validator"] = ValidatorsWrapper(
            validators, schema=schema, request=request, mode=mode
        )

    preparers = prop.metadata.get("preparers", None)
    if preparers:
        params["preparer"] = PreparersWrapper(
            preparers, schema=schema, request=request, mode=mode
        )

    title = prop.metadata.get("title", None)

    if title:
        params["title"] = title

    description = prop.metadata.get("description", None)
    if description:
        params["description"] = description

    params.update(kwargs)
    return params


class String(colander.String):
    def __init__(self, encoding=None, allow_empty=True):
        super().__init__(encoding=encoding, allow_empty=allow_empty)

    def serialize(self, node, appstruct):
        if appstruct is None:
            return colander.null
        return super().serialize(node, appstruct)

    def deserialize(self, node, cstruct):
        res = super().deserialize(node, cstruct)
        return res


class Mapping(colander.Mapping):
    def serialize(self, node, appstruct):
        if appstruct is None:
            appstruct = {}
        return super().serialize(node, appstruct)


class Boolean(colander.Boolean):
    def serialize(self, node, appstruct):
        if appstruct is None:
            return colander.null
        return super().serialize(node, appstruct)

    def deserialize(self, node, appstruct):
        if appstruct is None:
            appstruct = colander.null
        return super().deserialize(node, appstruct)


def dataclass_field_to_colander_schemanode(
    prop: dataclasses.Field,
    schema,
    request,
    oid_prefix="deformField",
    mode=None,
    default_tzinfo=pytz.UTC,
    metadata=None,
) -> colander.SchemaNode:

    """
    Converts ``dataclass.Field`` to ``colander.SchemaNode``

    :param prop: ``dataclass.Field`` object
    :param schema: ``dataclass`` class
    :param request: request object
    :param oid_prefix: prefix for field OID, defaults to 'deformField'
    :param mode: One of the following: ``'default'``, ``'edit'``, ``'edit-process'``
    :param default_tzinfo: Default timezone to use for ``datetime`` handling, defaults to ``pytz.UTC``
    :param metadata: additional metadata override

    :return: converted ``colander.SchemaNode``
    """
    metadata = metadata or {}

    t = dataclass_get_type(prop)
    t["metadata"].update(metadata)
    field_factory = t["metadata"].get("colander.field_factory", None)
    if field_factory:
        params = colander_params(
            prop,
            oid_prefix,
            typ=field_factory(request),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == date:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.Date(),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == datetime:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.DateTime(default_tzinfo=default_tzinfo),
            schema=schema,
            request=request,
            mode=mode,
        )

        return SchemaNode(**params)
    if t["type"] == str:
        params = colander_params(
            prop,
            oid_prefix,
            typ=String(allow_empty=True),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == int:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.Integer(),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == float:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.Float(),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == bool:
        params = colander_params(
            prop, oid_prefix, typ=Boolean(), schema=schema, request=request, mode=mode,
        )
        return SchemaNode(**params)

    if is_dataclass_field(prop):
        subtype = dc2colander(
            t["type"],
            request=request,
            colander_schema_type=colander.MappingSchema,
            mode=mode,
        )

        return subtype()
    if t["type"] == dict:
        params = colander_params(
            prop,
            oid_prefix,
            typ=Mapping(unknown="preserve"),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == list:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.List(),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == deform.FileData:
        params = colander_params(
            prop,
            oid_prefix,
            typ=deform.FileData(),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    if t["type"] == set:
        params = colander_params(
            prop,
            oid_prefix,
            typ=colander.Set(),
            schema=schema,
            request=request,
            mode=mode,
        )
        return SchemaNode(**params)
    raise KeyError(prop)


def dc2colander(
    schema: type,
    *,
    request: typing.Any = None,
    mode="default",
    include_fields: typing.List[str] = None,
    exclude_fields: typing.List[str] = None,
    hidden_fields: typing.List[str] = None,
    readonly_fields: typing.List[str] = None,
    include_schema_validators: bool = True,
    colander_schema_type: typing.Type[colander.Schema] = colander.MappingSchema,
    oid_prefix: str = "deformField",
    default_tzinfo=pytz.UTC,
    field_metadata=None,
    dataclass_field_to_colander_schemanode=dataclass_field_to_colander_schemanode,
) -> typing.Type[colander.MappingSchema]:
    """
    Converts ``dataclass`` to ``colander.Schema``

    :param schema: ``dataclass`` class to be used as schema
    :param request: a request object. This is mainly be passed down to downstream factories, accepts anything.
    :param mode: this flag is used to affect the decision on how validators validate data. accepts one of the following:
                 - ``'default'``
                 - ``'edit'`` - flag used when generating edit form
                 - ``'edit-process'`` - flag used when generating edit processing form
    :param include_fields: List of field names to include
    :type include_fields: typing.List[str]
    :param exclude_fields: List of field names to exclude
    :type exclude_fields: typing.List[str]
    :param hidden_fields: List of field names to hide
    :type hidden_fields: typing.List[str]
    :param readonly_fields: List of field names to made readonly
    :type readonly_fields: typing.List[str]
    :param include_schema_validators: Set whether to include ``__validators__`` from ``dataclass`` during convertion
    :type include_schema_validators: bool
    :param colander_schema_type: base class to use as created output, defaults to colander.MappingSchema.
    :param oid_prefix: string to use as deform OID prefix
    :param default_tzinfo: default timezone for ``datetime`` handling, defaults to ``pytz.UTC``
    :param field_metadata: a dictionary for overriding field metadata. Structure: ``{'<fieldname>': {'metadatakey': 'metadataval'}}``
    :param dataclass_field_to_colander_schemanode: ``colander.SchemaNode`` factory function.

    :return: ``colander.Schema`` class

    **Field metadata handling**

    This function will read several metadata from fields and use it to derive the parameters.

    - ``required: bool`` - flag field as required
    - ``title: str`` - field title
    - ``description: str`` - field description
    - ``validators: typing.List[typing.Callable]`` - a list of validator callables
    - ``preparers: typing.List[typing.Callable]`` - a list of preparer callables
    - ``deform.widget: deform.widget.Widget`` - ``deform.widget.Widget`` object to use as widget.
    - ``deform.widget_factory: typing.Callable`` - a callable with that accept ``request`` and
      returns a ``deform.widget.Widget`` object.
    - ``colander.field_factory: typing.Callable`` - a callable that accept ``request`` and returns a
      ``colander.SchemaType`` object.


    """
    # output colander schema from dataclass schema
    attrs = {}

    include_fields = include_fields or []
    exclude_fields = exclude_fields or []
    hidden_fields = hidden_fields or []
    readonly_fields = readonly_fields or []
    field_metadata = field_metadata or {}
    if mode == "edit":
        readonly_fields += [
            attr
            for attr, prop in schema.__dataclass_fields__.items()
            if (
                not prop.metadata.get("editable", True)
                or prop.metadata.get("readonly", False)
            )
        ]
    elif mode == "edit-process":
        exclude_fields += [
            attr
            for attr, prop in schema.__dataclass_fields__.items()
            if (
                not prop.metadata.get("editable", True)
                or prop.metadata.get("readonly", False)
            )
        ]
    else:
        readonly_fields += [
            attr
            for attr, prop in schema.__dataclass_fields__.items()
            if (prop.metadata.get("readonly", False))
        ]

    if include_fields:
        for attr, prop in schema.__dataclass_fields__.items():
            if prop.name in include_fields and prop.name not in exclude_fields:
                prop = dataclass_field_to_colander_schemanode(
                    prop,
                    oid_prefix=oid_prefix,
                    schema=schema,
                    request=request,
                    mode=mode,
                    metadata=field_metadata.get(prop.name, {}),
                    default_tzinfo=default_tzinfo,
                )
                attrs[attr] = prop
    else:
        for attr, prop in schema.__dataclass_fields__.items():
            if prop.name not in exclude_fields:
                prop = dataclass_field_to_colander_schemanode(
                    prop,
                    oid_prefix=oid_prefix,
                    schema=schema,
                    request=request,
                    mode=mode,
                    metadata=field_metadata.get(prop.name, {}),
                    default_tzinfo=default_tzinfo,
                )
                attrs[attr] = prop

    for attr, prop in attrs.items():
        dcprop = schema.__dataclass_fields__[attr]

        t = dataclass_get_type(dcprop)
        if attr in hidden_fields:
            if prop.widget is None:
                prop.widget = HiddenWidget()
            else:
                prop.widget.hidden = True

        if attr in readonly_fields:
            if prop.widget is None:
                prop_widget = default_widget_makers.get(prop.typ.__class__, None)
                if prop_widget is None:
                    prop_widget = TextInputWidget
                prop.widget = prop_widget()

            prop.widget.readonly = True

        if t["type"] == str:
            if dcprop.metadata.get("format", None) == "text":
                if prop.widget is None:
                    prop.widget = TextAreaWidget()

    if include_schema_validators:

        def validator(self, node, appstruct):
            vdata = replace_colander_null(appstruct)
            form_validators = copy.deepcopy(getattr(schema, "__validators__", []))
            # FIXME: this create a coupling with morpfw, need to decouple
            app = getattr(request, "app", None)
            if app:
                get_formvalidators = getattr(app, "get_formvalidators", None)
                if get_formvalidators:
                    form_validators += get_formvalidators(schema)

            for form_validator in form_validators:
                required_binds = getattr(form_validator, "__required_binds__", [])
                kwargs = {}
                for k in required_binds:
                    if self.bindings is None or (k not in self.bindings.keys()):
                        raise AssertionError(
                            "Required bind variable '{}' is not set on '{}'".format(
                                k, self
                            )
                        )
                    else:
                        kwargs[k] = self.bindings[k]

                fe = form_validator(
                    schema=schema,
                    data=vdata,
                    mode=mode,
                    **(self.bindings or {"request": request}),
                )
                if fe:
                    exc = colander.Invalid(node, fe["message"])
                    if fe.get("field", None):
                        exc[fe["field"]] = fe["message"]
                    if fe.get("fields", None):
                        for fn in fe["fields"]:
                            exc[fn] = fe["message"]
                    raise exc

        attrs["validator"] = validator

    Schema = type("Schema", (colander_schema_type,), attrs)

    return Schema


convert = dc2colander
